from datetime import date
import os
import shutil

import toml

import audeer


config = toml.load(audeer.path('..', 'pyproject.toml'))


# Project -----------------------------------------------------------------
author = ', '.join(author['name'] for author in config['project']['authors'])
copyright = f'2021-{date.today().year} audEERING GmbH'
project = config['project']['name']
version = audeer.git_repo_version()
title = 'Documentation'


# General -----------------------------------------------------------------
master_doc = 'index'
source_suffix = '.rst'
exclude_patterns = [
    'api-src',
    'build',
    'tests',
    'Thumbs.db',
    '.DS_Store',
]
templates_path = ['_templates']
pygments_style = None
extensions = [
    'jupyter_sphinx',  # executing code blocks
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',  # support for Google-style docstrings
    'sphinx.ext.autosummary',
    'sphinx.ext.viewcode',
    'sphinx.ext.intersphinx',
    'sphinx.ext.autosectionlabel',
    'sphinx_autodoc_typehints',
    'sphinx_copybutton',  # for "copy to clipboard" buttons
]

# Ignore package dependencies during building the docs
autodoc_mock_imports = [
    'tqdm',
]

# Reference with :ref:`data-header:Database`
autosectionlabel_prefix_document = True
autosectionlabel_maxdepth = 2

# Do not copy prompot output
copybutton_prompt_text = r'>>> |\.\.\. '
copybutton_prompt_is_regexp = True

# Mapping to external documentation
intersphinx_mapping = {
    'audobject': ('https://audeering.github.io/audobject/', None),
    'python': ('https://docs.python.org/3/', None),
    'opensmile': ('https://audeering.github.io/opensmile-python/', None),
    'torch': ('https://pytorch.org/docs/stable/', None),
}

# Disable Gitlab as we need to sign in
linkcheck_ignore = [
    'https://gitlab.audeering.com',
    'https://sphinx-doc.org',
]

# Disable auto-generation of TOC entries in the API
# https://github.com/sphinx-doc/sphinx/issues/6316
toc_object_entries = False


# HTML --------------------------------------------------------------------
html_theme = 'sphinx_audeering_theme'
html_theme_options = {
    'display_version': True,
    'logo_only': False,
    'footer_links': False,
}
html_context = {
    'display_github': True,
}
html_title = title
html_static_path = ['_static']


# Copy API (sub-)module RST files to docs/api/ folder ---------------------
audeer.rmdir('api')
audeer.mkdir('api')
api_src_files = audeer.list_file_names('api-src')
api_dst_files = [
    audeer.path('api', os.path.basename(src_file))
    for src_file in api_src_files
]
for src_file, dst_file in zip(api_src_files, api_dst_files):
    shutil.copyfile(src_file, dst_file)



import glob
import os
import shutil

import pytest

import audeer
import audiofile
import opensmile


pytest.ROOT = audeer.path(
    os.path.dirname(os.path.realpath(__file__))
)
pytest.TMP = audeer.mkdir(
    os.path.join(
        pytest.ROOT,
        audeer.uid(),
    )
)
pytest.SIGNAL, pytest.SAMPLING_RATE = audiofile.read(
    os.path.join(pytest.ROOT, 'test.wav'),
    always_2d=True,
)

# feature extractor

pytest.FEATURE = opensmile.Smile(
    feature_set=opensmile.FeatureSet.GeMAPSv01b,
    feature_level=opensmile.FeatureLevel.LowLevelDescriptors,
)
pytest.FEATURE_SHAPE = pytest.FEATURE(
    pytest.SIGNAL,
    pytest.SAMPLING_RATE,
).shape[1:]


# clean up

@pytest.fixture(scope='session', autouse=True)
def cleanup_session():
    path = os.path.join(
        pytest.TMP,
        '..',
        '.coverage.*',
    )
    for file in glob.glob(path):
        os.remove(file)
    yield
    if os.path.exists(pytest.TMP):
        shutil.rmtree(pytest.TMP)



import pytest

import audonnx


def mean(x, sr):
    return x.mean(axis=1)


@pytest.mark.parametrize(
    'node, expected',
    [
        (
            audonnx.InputNode(
                [1, -1],
                'tensor(float)',
                None,
            ),
            "{shape: [1, -1], dtype: tensor(float), transform: None}"
        ),
        (
            audonnx.InputNode(
                [pytest.FEATURE_SHAPE[0], -1],
                'tensor(float)',
                pytest.FEATURE,
            ),
            "{shape: [18, -1], dtype: tensor(float), "
            "transform: opensmile.core.smile.Smile}"
        ),
        (
            audonnx.InputNode(
                [1],
                'tensor(float)',
                audonnx.Function(lambda x, sr: x.mean(axis=1)),
            ),
            "{shape: [1], dtype: tensor(float), "
            "transform: audonnx.core.function.Function(<lambda>)}"
        ),
        (
            audonnx.InputNode(
                [1],
                'tensor(float)',
                audonnx.Function(mean),
            ),
            "{shape: [1], dtype: tensor(float), "
            "transform: audonnx.core.function.Function(mean)}"
        ),
        (
            audonnx.OutputNode(
                [1, 2],
                'tensor(float)',
                ['female', 'male'],
            ),
            "{shape: [1, 2], dtype: tensor(float), labels: [female, male]}"
        ),
    ],
)
def test_str(node, expected):
    assert str(node) == expected



import os
import re
import typing

import numpy as np
import onnx
import onnxruntime
import yaml

import audeer
import audobject

from audonnx.core.node import InputNode
from audonnx.core.node import OutputNode
from audonnx.core.typing import Device
from audonnx.core.typing import Labels
from audonnx.core.typing import Transform


class Model(audobject.Object):
    r"""Model with multiple input and output nodes.

    For input nodes an optional transform can be given
    that transforms the raw signal into the desired representation,
    otherwise the raw signal is used as input.
    Use dictionary to assign transform objects to specific nodes
    if model has multiple input nodes.

    For output nodes an optional list of labels can be given
    to assign names to the last non-dynamic dimension.
    E.g. if the shape of the output node is
    ``(1, 3, -1)``
    three labels can be assigned to the second dimension.
    By default,
    labels are generated from the name of the node.
    Use dictionary to assign labels to specific nodes
    if model has multiple output nodes.

    :class:`Model` inherites from :class:`audobject.Object`,
    which means you can serialize to
    and instantiate the class
    from a YAML file.
    Have a look at :class:`audobject.Object`
    to see all available methods.

    Args:
        path: ONNX proto object or path to ONNX file
        labels: list of labels or dictionary with labels
        transform: callable object or a dictionary of callable objects
        device: set device
            (``'cpu'``, ``'cuda'``, or ``'cuda:<id>'``)
            or a (list of) provider(s)_

    Examples:
        >>> import audiofile
        >>> import opensmile
        >>> transform = opensmile.Smile(
        ...    opensmile.FeatureSet.GeMAPSv01b,
        ...    opensmile.FeatureLevel.LowLevelDescriptors,
        ... )
        >>> path = os.path.join('tests', 'model.onnx')
        >>> model = Model(
        ...     path,
        ...     labels=['female', 'male'],
        ...     transform=transform,
        ... )
        >>> model
        Input:
          feature:
            shape: [18, -1]
            dtype: tensor(float)
            transform: opensmile.core.smile.Smile
        Output:
          gender:
            shape: [2]
            dtype: tensor(float)
            labels: [female, male]
        >>> path = os.path.join('tests', 'test.wav')
        >>> signal, sampling_rate = audiofile.read(path)
        >>> model(
        ...     signal,
        ...     sampling_rate,
        ...     outputs='gender',
        ... ).round(1)
        array([-195.1,   73.3], dtype=float32)

    .. _provider(s): https://onnxruntime.ai/docs/execution-providers/

    """
    @audobject.init_decorator(
        borrow={
            'labels': '_original_args',
            'transform': '_original_args',
        },
        resolvers={
            'path': audobject.resolver.FilePath,
        },
        hide=[
            'device',
        ],
    )
    def __init__(
            self,
            path: typing.Union[str, onnx.ModelProto],
            *,
            labels: Labels = None,
            transform: Transform = None,
            device: Device = 'cpu',
    ):
        # keep original arguments to store them
        # when object is serialized
        self._original_args = {
            'labels': labels,
            'transform': transform,
        }

        self.path = audeer.path(path) if isinstance(path, str) else None
        r"""Model path"""

        providers = _device_to_providers(device)
        self.sess = onnxruntime.InferenceSession(
            self.path if isinstance(path, str) else path.SerializeToString(),
            providers=providers,
        )
        r"""Interference session"""

        inputs = self.sess.get_inputs()
        outputs = self.sess.get_outputs()

        transform = transform or {}
        if not isinstance(transform, dict):
            if not len(inputs) == 1:
                names = [input.name for input in inputs]
                raise ValueError(
                    f'Model has multiple input nodes. '
                    f'Please use a dictionary to assign '
                    f'transform object to one of '
                    f'{names}.'
                )
            transform = {inputs[0].name: transform}

        labels = labels or {}
        if not isinstance(labels, dict):
            if not len(outputs) == 1:
                names = [output.name for output in outputs]
                raise ValueError(
                    f'Model has multiple output nodes. '
                    f'Please use a dictionary to assign '
                    f'labels to one of '
                    f'{names}.'
                )
            labels = {outputs[0].name: labels}

        self.inputs = {}
        r"""Input nodes"""
        for input in inputs:
            trans = transform[input.name] if input.name in transform else None
            shape = _shape(input.shape)
            self.inputs[str(input.name)] = InputNode(
                shape,
                input.type,
                trans,
            )

        self.outputs = {}
        r"""Output nodes"""
        for output in outputs:
            shape = output.shape or [1]
            shape = _shape(shape)
            dim_size = _last_static_dim_size(shape)
            if output.name in labels:
                lab = labels[output.name]
                if len(lab) != dim_size:
                    raise ValueError(
                        f"Cannot assign "
                        f"{len(lab)} "
                        f"labels to output node "
                        f"'{output.name}' "
                        f"when last non-dynamic dimension has size "
                        f"{dim_size}."
                    )
            elif dim_size == 1:
                lab = [output.name]
            else:
                lab = [f'{output.name}-{idx}'
                       for idx in range(dim_size)]
            self.outputs[output.name] = OutputNode(
                shape,
                output.type,
                lab,
            )

    @audeer.deprecated_keyword_argument(
        deprecated_argument='output_names',
        removal_version='1.2.0',
        new_argument='outputs',
    )
    def __call__(
            self,
            signal: np.ndarray,
            sampling_rate: int,
            *,
            outputs: typing.Union[str, typing.Sequence[str]] = None,
            concat: bool = False,
            squeeze: bool = False,
    ) -> typing.Union[
        np.ndarray,
        typing.Dict[str, np.ndarray],
    ]:
        r"""Compute output for one or more nodes.

        If ``outputs`` is a plain string,
        the output of the according node is returned.

        If ``outputs`` is a list of strings,
        a dictionary with according nodes as keys and
        their outputs as values is returned.

        If ``outputs`` is not set
        and the model has a single output node,
        the output of that node is returned.
        Otherwise a dictionary with outputs of all nodes is returned.

        If ``concat`` is set to ``True``,
        the output of the requested nodes is concatenated
        along the last non-dynamic axis
        and a single array is returned.
        This requires that the number of dimensions,
        the position of dynamic axis,
        and all dimensions except the last non-dynamic axis
        match for the requested nodes

        Use :attr:`audonnx.Model.outputs` to get a list of available
        output nodes.

        Args:
            signal: input signal
            sampling_rate: sampling rate in Hz
            outputs: name of output or list with output names
            concat: if ``True``,
                concatenate output of the requested nodes
            squeeze: if ``True``,
                remove axes of length one from the output(s)

        Returns:
            model output

        Raises:
            RuntimeError: if ``concat`` is ``True``,
                but output of requested nodes cannot be concatenated

        """
        if outputs is None:
            outputs = list(self.outputs)
            if len(outputs) == 1:
                outputs = outputs[0]

        y = {}
        for name, input in self.inputs.items():
            if input.transform is not None:
                x = input.transform(signal, sampling_rate)
            else:
                x = signal
            y[name] = x.reshape(self.inputs[name].shape)

        z = self.sess.run(
            audeer.to_list(outputs),
            y,
        )

        if isinstance(outputs, str):
            z = z[0]
        else:
            z = {
                name: values for name, values in zip(outputs, z)
            }
            if concat:
                shapes = [self.outputs[node].shape for node in outputs]
                z = _concat(z, shapes)

        if squeeze:
            if isinstance(z, dict):
                z = {
                    name: values.squeeze() for name, values in z.items()
                }
            else:
                z = z.squeeze()

        return z

    def __repr__(self) -> str:
        r"""Printable representation of model."""
        d = {
            'Input': {
                name: node._dict() for name, node in self.inputs.items()
            },
            'Output': {
                name: node._dict() for name, node in self.outputs.items()
            }
        }
        return yaml.dump(d, default_flow_style=None).strip()

    def __str__(self) -> str:
        r"""String representation of model."""
        return repr(self)

    def labels(
            self,
            outputs: typing.Union[str, typing.Sequence[str]] = None,
    ) -> typing.Sequence[str]:
        r"""Collect labels of output nodes.

        Args:
            outputs: name of output or list with output names.
                Selects all output nodes by default

        Returns:
            list with labels

        """
        if outputs is None:
            outputs = list(self.outputs)
        outputs = audeer.to_list(outputs)

        names = [self.outputs[name].labels for name in outputs]
        names = audeer.flatten_list(names)

        return names

    def to_yaml(
            self,
            path: str,
            *,
            include_version: bool = True,
    ):
        r"""Save model to YAML file.

        If ONNX model was loaded from a byte stream,
        it will be saved in addition to the YAML file
        under the same path
        with file extension ``.onnx``.

        Args:
            path: file path, must end on ``.yaml``
            include_version: add version to class name

        Raises:
            ValueError: if file path does not end on ``.yaml``

        """
        path = audeer.path(path)
        if not audeer.file_extension(path) == 'yaml':
            raise ValueError(f"Model path {path} does not end on '.yaml'")

        if self.path is None:
            # if model was loaded from a byte stream,
            # we have to save it first
            self.path = audeer.replace_file_extension(path, 'onnx')
            audeer.mkdir(os.path.dirname(path))
            onnx.save(self.sess._model_bytes, self.path)

        with open(path, 'w') as fp:
            super().to_yaml(fp, include_version=include_version)


def _device_to_providers(
        device: typing.Union[
            str,
            typing.Tuple[str, typing.Dict],
            typing.Sequence[typing.Union[str, typing.Tuple[str, typing.Dict]]],
        ],
) -> typing.Sequence[typing.Union[str, typing.Tuple[str, typing.Dict]]]:
    r"""Converts device into a list of providers."""
    if isinstance(device, str):
        if device == 'cpu':
            providers = ['CPUExecutionProvider']
        elif device.startswith('cuda'):
            match = re.search(r'^cuda:(\d+)$', device)
            if match:
                device_id = match.group(1)
                providers = [
                    (
                        'CUDAExecutionProvider', {
                            'device_id': device_id,
                        }
                    ),
                ]
            else:
                providers = ['CUDAExecutionProvider']
        else:
            providers = [device]
    elif isinstance(device, tuple):
        providers = [device]
    else:
        providers = device
    return providers


def _concat(
        y: typing.Dict[str, np.ndarray],
        shapes: typing.Sequence[typing.List[int]],
):
    r"""Flatten dictionary by concatenating values."""
    y = list(y.values())

    # special case if all shapes are [-1]
    if all(map([-1].__eq__, shapes)):
        return np.stack(y)

    axis = _concat_axis(shapes)
    if axis is None:
        raise RuntimeError(
            f'To concatenate outputs '
            f'number of dimensions, '
            f'position of dynamic axis, '
            f'and all dimensions except the last non-dynamic axis '
            f'must match. '
            f'This does not apply to: '
            f'{shapes}'
        )

    return np.concatenate(y, axis=axis)


def _concat_axis(shapes: typing.Sequence[int]) -> typing.Optional[int]:
    r"""Return concat dimension or None if not possible."""
    # number of dimensions do not match
    if not len(set(map(len, shapes))) == 1:
        return None

    # dynamic axis in different positions
    if not len(set(map(_dynamic_axis, shapes))) == 1:
        return None

    # select last non-dynamic axis
    axis = len(shapes[0]) - (2 if shapes[0][-1] == -1 else 1)

    # remove concat axis
    shapes_wo_axis = [shape[:axis] + shape[axis + 1:] for shape in shapes]

    # non-concat dimensions do not match
    if not all(map(shapes_wo_axis[0].__eq__, shapes_wo_axis)):
        return None

    return axis


def _dynamic_axis(shape: typing.Sequence[int]) -> typing.Optional[int]:
    r"""Return dimension of dynamic axis or None if none."""
    for idx, dim in enumerate(shape):
        if dim == -1:
            return idx
    return None


def _last_static_dim_size(
        shape: typing.List[int],
) -> int:
    r"""Return size of last static dimension."""
    shape = list(filter((-1).__ne__, shape))
    return shape[-1] if len(shape) else 0


def _shape(
        shape: typing.List[typing.Union[int, str]],
) -> typing.List[int]:
    r"""Replace dynamic dimensions with -1."""
    return [-1 if not isinstance(x, int) else x for x in shape]



import typing

import numpy as np  # noqa: F401, needed for doctest
import onnx

from audonnx.core.function import Function
from audonnx.core.model import Model
from audonnx.core.typing import Device


def create_model(
        shapes: typing.Sequence[typing.Sequence[int]],
        *,
        value: float = 0.,
        dtype: int = onnx.TensorProto.FLOAT,
        opset_version: int = 14,
        device: Device = 'cpu',
) -> Model:
    r"""Create test model.

    Creates a model that outputs
    arrays filled with ``value``
    of the given ``shapes``.
    For each entry an output node will be created.
    ``-1``, ``None`` or strings
    define a dynamic axis.
    Per node,
    one dynamic axis is allowed.

    Args:
        shapes: list with shapes defining the identical
            input and output nodes of the model graph
        value: fill value
        dtype: data type, see `supported data types`_
        opset_version: opset version
        device: set device
            (``'cpu'``, ``'cuda'``, or ``'cuda:<id>'``)
            or a (list of) provider(s)_

    Returns:
        model object

    Examples:
        >>> shapes = [[3], [1, -1, 2]]
        >>> model = create_model(shapes)
        >>> model
        Input:
          input-0:
            shape: [3]
            dtype: tensor(float)
            transform: audonnx.core.function.Function(reshape)
          input-1:
            shape: [1, -1, 2]
            dtype: tensor(float)
            transform: audonnx.core.function.Function(reshape)
        Output:
          output-0:
            shape: [3]
            dtype: tensor(float)
            labels: [output-0-0, output-0-1, output-0-2]
          output-1:
            shape: [1, -1, 2]
            dtype: tensor(float)
            labels: [output-1-0, output-1-1]
        >>> signal = np.zeros((1, 5), np.float32)
        >>> model(signal, 8000)
        {'output-0': array([0., 0., 0.], dtype=float32), 'output-1': array([[[0., 0.],
                [0., 0.],
                [0., 0.],
                [0., 0.],
                [0., 0.]]], dtype=float32)}

    .. _`supported data types`: https://onnxruntime.ai/docs/reference/operators/custom-python-operator.html#supported-data-types

    """  # noqa: E501
    # create graph

    object = create_model_proto(
        shapes,
        dtype=dtype,
        opset_version=opset_version,
    )

    # create transform objects

    transform = {}
    for idx, shape in enumerate(shapes):
        transform[f'input-{idx}'] = Function(
            reshape,
            func_args={
                'shape': shape,
                'value': value,
            },
        )

    # create model

    model = Model(
        object,
        transform=transform,
        device=device,
    )

    return model


def create_model_proto(
        shapes: typing.Sequence[typing.Sequence[int]],
        *,
        dtype: int = onnx.TensorProto.FLOAT,
        opset_version: int = 14,
) -> onnx.ModelProto:
    r"""Create test ONNX proto object.

    Creates an identity graph
    with input and output nodes
    of the given ``shapes``.
    For each entry an input and output
    node will be created.
    ``-1``, ``None`` or strings
    define a dynamic axis.
    Per node,
    one dynamic axis is allowed.
    The identity graph can be used as ``path`` argument
    of :class:`audonnx.Model`.

    Args:
        shapes: list with shapes defining the identical
            input and output nodes of the model graph
        dtype: data type, see `supported data types`_
        opset_version: opset version

    Returns:
        ONNX object

    Examples:
        >>> create_model_proto([[2]])
        ir_version: 9
        opset_import {
          version: 14
        }
        producer_name: "test"
        graph {
          node {
            input: "input-0"
            output: "output-0"
            op_type: "Identity"
          }
          name: "test"
          input {
            name: "input-0"
            type {
              tensor_type {
                elem_type: 1
                shape {
                  dim {
                    dim_value: 2
                  }
                }
              }
            }
          }
          output {
            name: "output-0"
            type {
              tensor_type {
                elem_type: 1
                shape {
                  dim {
                    dim_value: 2
                  }
                }
              }
            }
          }
        }
        <BLANKLINE>

    .. _`supported data types`: https://onnxruntime.ai/docs/reference/operators/custom-python-operator.html#supported-data-types

    """  # noqa: E501
    # nodes

    inputs = []
    outputs = []
    nodes = []

    for idx, shape in enumerate(shapes):
        input = onnx.helper.make_tensor_value_info(
            f'input-{idx}',
            dtype,
            shape,
        )
        inputs.append(input)

        output = onnx.helper.make_tensor_value_info(
            f'output-{idx}',
            dtype,
            shape,
        )
        outputs.append(output)

        node = onnx.helper.make_node(
            'Identity',
            [input.name],
            [output.name],
        )
        nodes.append(node)

    # graph

    graph = onnx.helper.make_graph(
        nodes,
        'test',
        inputs,
        outputs,
    )

    # model

    model = onnx.helper.make_model(graph, producer_name='test')
    model.opset_import[0].version = opset_version
    onnx.checker.check_model(model)

    return model


def reshape(signal, _, shape, value):
    r"""Return array with zeros of given shape."""
    import numpy as np  # noqa: F811
    shape = [signal.shape[-1] if not isinstance(s, int) or s < 0
             else s for s in shape]
    return (np.ones(shape) * value).astype(signal.dtype)



import os
import subprocess
import sys

import pkg_resources
import pytest

import audonnx.testing


def uninstall(
    package: str,
    module: str,
):
    # uninstall package
    subprocess.check_call(
        [
            sys.executable,
            '-m',
            'pip',
            'uninstall',
            '--yes',
            package,
        ]
    )
    # remove module
    for m in list(sys.modules):
        if m.startswith(package):
            sys.modules.pop(m)
    # force pkg_resources to re-scan site packages
    pkg_resources._initialize_master_working_set()


def test(tmpdir):

    object = audonnx.testing.create_model_proto([pytest.FEATURE_SHAPE])
    model = audonnx.Model(
        object,
        transform=pytest.FEATURE,
    )
    model_path = os.path.join(tmpdir, 'model.yaml')
    model.to_yaml(model_path)

    # Removing the package does not work under Windows
    # as the DLL file cannot be unloaded
    if not sys.platform == 'win32':
        uninstall('opensmile', 'opensmile')

        with pytest.raises(ModuleNotFoundError):
            audonnx.load(tmpdir)

    audonnx.load(tmpdir, auto_install=True)



import os
import typing

import oyaml as yaml

import audeer
import audobject

from audonnx.core.model import Model


def load(
        root: str,
        *,
        model_file: str = 'model.yaml',
        labels_file: str = 'labels.yaml',
        transform_file: str = 'transform.yaml',
        device: typing.Union[
            str,
            typing.Tuple[str, typing.Dict],
            typing.Sequence[typing.Union[str, typing.Tuple[str, typing.Dict]]],
        ] = 'cpu',
        auto_install: bool = False,
) -> Model:
    r"""Load model from folder.

    Tries to load model from YAML file.

    Otherwise creates object from ONNX file (legacy mode).
    In this case it will load
    labels and the transform
    from the corresponding YAML files
    if provided.
    The model is expected to be located at
    ``audeer.replace_file_extension(model_file, 'onnx')``.

    Args:
        root: root folder
        model_file: model YAML file,
            that needs to end with ``.yaml``.
            In legacy mode path to model ONNX file
        labels_file: YAML file with labels
        transform_file: YAML file with transformation
        device: set device
            (``'cpu'``, ``'cuda'``, or ``'cuda:<id>'``)
            or a (list of) provider_
        auto_install: install missing packages needed to create the object

    .. _provider: https://onnxruntime.ai/docs/execution-providers/

    Returns:
        model

    Examples:
        >>> model = load('tests')
        >>> model
        Input:
          feature:
            shape: [18, -1]
            dtype: tensor(float)
            transform: opensmile.core.smile.Smile
        Output:
          gender:
            shape: [2]
            dtype: tensor(float)
            labels: [female, male]

    """  # noqa: E501
    root = audeer.path(root)
    model_file = os.path.join(root, model_file)
    model_file_yaml = audeer.replace_file_extension(model_file, 'yaml')
    if audeer.file_extension(model_file) == 'yaml':
        model_file_onnx = audeer.replace_file_extension(model_file, 'onnx')
    else:
        model_file_onnx = model_file

    # Try to load object from YAML file

    if os.path.exists(model_file_yaml):
        with open(model_file_yaml) as f:
            first_line = f.readline()
        if first_line.startswith('$audonnx'):  # ensure correct object
            return audobject.from_yaml(
                model_file_yaml,
                auto_install=auto_install,
                override_args={
                    'device': device,
                },
            )

    # LEGACY support
    # Otherwise create object from ONNX file

    labels_file = os.path.join(root, labels_file)
    transform_file = os.path.join(root, transform_file)

    labels = None
    if os.path.exists(labels_file):
        with open(labels_file, 'r') as fp:
            labels = yaml.load(fp, yaml.BaseLoader)

    transform = None
    if os.path.exists(transform_file):
        transform = audobject.from_yaml(
            transform_file,
            auto_install=auto_install,
        )

    model = Model(
        model_file_onnx,
        labels=labels,
        transform=transform,
        device=device,
    )

    return model
